@@ -44,6 +44,10 @@
 #include <unistd.h>
 #include <getopt.h>
 
+#define KB (1024)
+#define MB (1024*KB)
+#define GB (1024*MB)
+
 #define REPORT_UTIL "hugeadm"
 #define REPORT(level, prefix, format, ...)                                    \
 	do {                                                                  \
@@ -79,6 +83,14 @@ extern char *optarg;
 #define SWAP_FREE "SwapFree:"
 #define SWAP_TOTAL "SwapTotal:"
 
+#define ALWAYS		  "always"
+#define MADVISE		  "madvise"
+#define NEVER		  "never"
+#define TRANS_ENABLE	  "/sys/kernel/mm/transparent_hugepage/enabled"
+#define KHUGE_SCAN_PAGES  "/sys/kernel/mm/transparent_hugepage/khugepaged/pages_to_scan"
+#define KHUGE_SCAN_SLEEP  "/sys/kernel/mm/transparent_hugepage/khugepaged/scan_sleep_millisecs"
+#define KHUGE_ALLOC_SLEEP "/sys/kernel/mm/transparent_hugepage/khugepaged/alloc_sleep_millisecs"
+
 void print_usage()
 {
 	fprintf(stderr, "hugeadm [options]\n");
@@ -93,6 +105,15 @@ void print_usage()
 	CONT("Adjust pool 'size' lower bound");
 	OPTION("--obey-mempolicy", "Obey the NUMA memory policy when");
 	CONT("adjusting the pool 'size' lower bound");
+	OPTION("--thp-always", "Enable transparent huge pages always");
+	OPTION("--thp-madvise", "Enable transparent huge pages with madvise");
+	OPTION("--thp-never", "Disable transparent huge pages");
+	OPTION("--thp-khugepaged-pages <pages to scan>", "Number of pages that khugepaged");
+	CONT("should scan on each pass");
+	OPTION("--thp-khugepaged-scan-sleep <milliseconds>", "Time in ms to sleep between");
+	CONT("khugepaged passes");
+	OPTION("--thp-khugepages-alloc-sleep <milliseconds>", "Time in ms for khugepaged");
+	CONT("to wait if there was a huge page allocation failure");
 	OPTION("--pool-pages-max <size|DEFAULT>:[+|-]<pagecount|memsize<G|M|K>>", "");
 	CONT("Adjust pool 'size' upper bound");
 	OPTION("--set-recommended-min_free_kbytes", "");
@@ -267,6 +288,16 @@ void verbose_expose(void)
 
 #define LONG_EXPLAIN	('e' << 8)
 
+#define LONG_TRANS			('t' << 8)
+#define LONG_TRANS_ALWAYS		(LONG_TRANS|'a')
+#define LONG_TRANS_MADVISE		(LONG_TRANS|'m')
+#define LONG_TRANS_NEVER		(LONG_TRANS|'n')
+
+#define LONG_KHUGE			('K' << 8)
+#define LONG_KHUGE_PAGES		(LONG_KHUGE|'p')
+#define LONG_KHUGE_SCAN			(LONG_KHUGE|'s')
+#define LONG_KHUGE_ALLOC		(LONG_KHUGE|'a')
+
 #define MAX_POOLS	32
 
 static int cmpsizes(const void *p1, const void *p2)
@@ -485,6 +516,9 @@ int mount_dir(char *path, char *options, mode_t mode, uid_t uid, gid_t gid)
 	struct group *grp;
 	struct mntent entry;
 	FILE *mounts;
+        char dummy;
+        int useMtab;
+
 	struct mount_list *list, *previous;
 
 	list = collect_active_mounts(NULL);
@@ -521,6 +555,14 @@ int mount_dir(char *path, char *options, mode_t mode, uid_t uid, gid_t gid)
 			return 1;
 		}
 
+          /* Check if mtab is a symlink */
+          useMtab = (readlink(MOUNTED, &dummy, 1) < 0);
+          if (!useMtab) {
+           /* No need updating mtab */
+              return 0;
+           }
+
+
 		mounts = setmntent(MOUNTED, "a+");
 		if (mounts) {
 			entry.mnt_fsname = FS_NAME;
@@ -553,6 +595,15 @@ int mount_dir(char *path, char *options, mode_t mode, uid_t uid, gid_t gid)
 	return 0;
 }
 
+void scale_size(char *buf, unsigned long pagesize)
+{
+	if(pagesize >= GB)
+		snprintf(buf, OPT_MAX, "%luGB", pagesize / GB);
+	else if(pagesize >= MB)
+		snprintf(buf, OPT_MAX, "%luMB", pagesize / MB);
+	else
+		snprintf(buf, OPT_MAX, "%luKB", pagesize / KB);
+}
 
 void create_mounts(char *user, char *group, char *base, mode_t mode)
 {
@@ -560,6 +611,7 @@ void create_mounts(char *user, char *group, char *base, mode_t mode)
 	char path[PATH_MAX];
 	char options[OPT_MAX];
 	char limits[OPT_MAX];
+	char scaled[OPT_MAX];
 	int cnt, pos;
 	struct passwd *pwd;
 	struct group *grp;
@@ -598,15 +650,17 @@ void create_mounts(char *user, char *group, char *base, mode_t mode)
 	}
 
 	for (pos=0; cnt--; pos++) {
+		scaled[0] = 0;
+		scale_size(scaled, pools[pos].pagesize);
 		if (user)
-			snprintf(path, PATH_MAX, "%s/%s/pagesize-%ld",
-				base, user, pools[pos].pagesize);
+			snprintf(path, PATH_MAX, "%s/%s/pagesize-%s",
+				base, user, scaled);
 		else if (group)
-			snprintf(path, PATH_MAX, "%s/%s/pagesize-%ld",
-				base, group, pools[pos].pagesize);
+			snprintf(path, PATH_MAX, "%s/%s/pagesize-%s",
+				base, group, scaled);
 		else
-			snprintf(path, PATH_MAX, "%s/pagesize-%ld",
-				base, pools[pos].pagesize);
+			snprintf(path, PATH_MAX, "%s/pagesize-%s",
+				base, scaled);
 
 		snprintf(options, OPT_MAX, "pagesize=%ld",
 				pools[pos].pagesize);
@@ -860,6 +914,13 @@ void check_user(void)
 	uid = getuid();
 	pwd = getpwuid(uid);
 
+	/* Don't segfault if user does not have a passwd entry. */
+	if (!pwd) {
+		printf("\n");
+		WARNING("User uid %d is not in the password file!\n", uid);
+		return;
+	}
+
 	if (gid != pwd->pw_gid && !user_in_group(grp->gr_mem, pwd->pw_name) && uid != 0) {
 		printf("\n");
 		WARNING("User %s (uid: %d) is not a member of the hugetlb_shm_group %s (gid: %d)!\n", pwd->pw_name, uid, grp->gr_name, gid);
@@ -1039,6 +1100,30 @@ void rem_ramdisk_swap(){
 	}
 }
 
+void set_trans_opt(const char *file, const char *value)
+{
+	FILE *f;
+
+	if (geteuid() != 0) {
+                ERROR("Transparent huge page options can only be set by root\n");
+                exit(EXIT_FAILURE);
+        }
+
+	if (opt_dry_run) {
+		printf("echo '%s' > %s\n", value, file);
+		return;
+	}
+
+	f = fopen(file, "w");
+	if (!f) {
+		ERROR("Couldn't open %s: %s\n", file, strerror(errno));
+		return;
+	}
+
+	fprintf(f, "%s", value);
+	fclose(f);
+}
+
 enum {
 	POOL_MIN,
 	POOL_MAX,
@@ -1285,7 +1370,10 @@ int main(int argc, char** argv)
 	int opt_list_mounts = 0, opt_pool_list = 0, opt_create_mounts = 0;
 	int opt_global_mounts = 0, opt_pgsizes = 0, opt_pgsizes_all = 0;
 	int opt_explain = 0, minadj_count = 0, maxadj_count = 0;
+	int opt_trans_always = 0, opt_trans_never = 0, opt_trans_madvise = 0;
+	int opt_khuge_pages = 0, opt_khuge_scan = 0, opt_khuge_alloc = 0;
 	int ret = 0, index = 0;
+	char *khuge_pages = NULL, *khuge_alloc = NULL, *khuge_scan = NULL;
 	gid_t opt_gid = 0;
 	struct group *opt_grp = NULL;
 	int group_invalid = 0;
@@ -1298,6 +1386,12 @@ int main(int argc, char** argv)
 		{"pool-pages-min", required_argument, NULL, LONG_POOL_MIN_ADJ},
 		{"pool-pages-max", required_argument, NULL, LONG_POOL_MAX_ADJ},
 		{"obey-mempolicy", no_argument, NULL, LONG_POOL_MEMPOL},
+		{"thp-always", no_argument, NULL, LONG_TRANS_ALWAYS},
+		{"thp-madvise", no_argument, NULL, LONG_TRANS_MADVISE},
+		{"thp-never", no_argument, NULL, LONG_TRANS_NEVER},
+		{"thp-khugepaged-pages", required_argument, NULL, LONG_KHUGE_PAGES},
+		{"thp-khugepaged-scan-sleep", required_argument, NULL, LONG_KHUGE_SCAN},
+		{"thp-khugepaged-alloc-sleep", required_argument, NULL, LONG_KHUGE_ALLOC},
 		{"set-recommended-min_free_kbytes", no_argument, NULL, LONG_SET_RECOMMENDED_MINFREEKBYTES},
 		{"set-recommended-shmmax", no_argument, NULL, LONG_SET_RECOMMENDED_SHMMAX},
 		{"set-shm-group", required_argument, NULL, LONG_SET_HUGETLB_SHM_GROUP},
@@ -1401,6 +1495,33 @@ int main(int argc, char** argv)
 			opt_obey_mempolicy = 1;
 			break;
 
+		case LONG_TRANS_ALWAYS:
+			opt_trans_always = 1;
+			break;
+
+		case LONG_TRANS_MADVISE:
+			opt_trans_madvise = 1;
+			break;
+
+		case LONG_TRANS_NEVER:
+			opt_trans_never = 1;
+			break;
+
+		case LONG_KHUGE_PAGES:
+			opt_khuge_pages = 1;
+			khuge_pages = optarg;
+			break;
+
+		case LONG_KHUGE_SCAN:
+			opt_khuge_scan = 1;
+			khuge_scan = optarg;
+			break;
+
+		case LONG_KHUGE_ALLOC:
+			opt_khuge_alloc = 1;
+			khuge_alloc = optarg;
+			break;
+
 		case LONG_POOL_MAX_ADJ:
 			if (! kernel_has_overcommit()) {
 				ERROR("kernel does not support overcommit, "
@@ -1511,6 +1632,24 @@ int main(int argc, char** argv)
 	if (opt_movable != -1)
 		setup_zone_movable(opt_movable);
 
+	if (opt_trans_always)
+		set_trans_opt(TRANS_ENABLE, ALWAYS);
+
+	if (opt_trans_madvise)
+		set_trans_opt(TRANS_ENABLE, MADVISE);
+
+	if (opt_trans_never)
+		set_trans_opt(TRANS_ENABLE, NEVER);
+
+	if (opt_khuge_pages)
+		set_trans_opt(KHUGE_SCAN_PAGES, khuge_pages);
+
+	if (opt_khuge_alloc)
+		set_trans_opt(KHUGE_ALLOC_SLEEP, khuge_alloc);
+
+	if (opt_khuge_scan)
+		set_trans_opt(KHUGE_SCAN_SLEEP, khuge_scan);
+
 	if (opt_set_recommended_minfreekbytes)
 		set_recommended_minfreekbytes();
 
