@@ -233,6 +233,52 @@ int file_write_ulong(char *file, unsigned long val)
 	return ret > 0 ? 0 : -1;
 }
 
+
+/*
+ * Return the name of this executable, using buf as temporary space.
+ */
+#define MAX_EXE 4096
+static char *get_exe_name(char *buf, int size)
+{
+	char *p;
+	int fd;
+	ssize_t nread;
+
+	buf[0] = 0;
+	fd = open("/proc/self/cmdline", O_RDONLY);
+	if (fd < 0) {
+		WARNING("Unable to open cmdline, no exe name\n");
+		return buf;
+	}
+	nread = read(fd, buf, size-1);
+	close(fd);
+
+	if (nread < 0) {
+		WARNING("Error %d reading cmdline, no exe name\n", errno);
+		return buf;
+	}
+	if (nread == 0) {
+		WARNING("Read zero bytes from cmdline, no exe name\n");
+		return buf;
+	}
+
+	buf[nread] = 0; /* make sure we're null terminated */
+	/*
+	 * Take advantage of cmdline being a series of null-terminated
+	 * strings.  The first string is the path to the executable in
+	 * the form:
+	 *
+	 *      /path/to/exe
+	 *
+	 * The exe name starts one character after the last '/'.
+	 */
+	p = strrchr(buf, '/');
+	if (!p)
+		return buf;
+	return p + 1;           /* skip over "/" */
+}
+
+
 /*
  * Reads the contents of hugetlb environment variables and save their
  * values for later use.
@@ -253,6 +299,29 @@ void hugetlbfs_setup_env()
 		__hugetlbfs_verbose = VERBOSE_DEBUG;
 	}
 
+	env = getenv("HUGETLB_RESTRICT_EXE");
+	if (env) {
+		char *p, *tok, *exe, buf[MAX_EXE+1], restrict[MAX_EXE];
+		int found = 0;
+
+		exe = get_exe_name(buf, sizeof buf);
+		DEBUG("Found HUGETLB_RESTRICT_EXE, this exe is \"%s\"\n", exe);
+		strncpy(restrict, env, sizeof restrict);
+		restrict[sizeof(restrict)-1] = 0;
+		for (p = restrict; (tok = strtok(p, ":")) != NULL; p = NULL) {
+			DEBUG("  ...check exe match for \"%s\"\n",  tok);
+			if (strcmp(tok, exe) == 0) {
+				found = 1;
+				DEBUG("exe match - libhugetlbfs is active for this exe\n");
+				break;
+			}
+		}
+		if (!found) {
+			DEBUG("No exe match - libhugetlbfs is inactive for this exe\n");
+			return;
+		}
+	}
+
 	env = getenv("HUGETLB_NO_PREFAULT");
 	if (env)
 		__hugetlbfs_prefault = false;
@@ -266,6 +335,15 @@ void hugetlbfs_setup_env()
 	__hugetlb_opts.morecore = getenv("HUGETLB_MORECORE");
 	__hugetlb_opts.heapbase = getenv("HUGETLB_MORECORE_HEAPBASE");
 
+	if (__hugetlb_opts.morecore)
+		__hugetlb_opts.thp_morecore =
+			(strcasecmp(__hugetlb_opts.morecore, "thp") == 0);
+
+	if (__hugetlb_opts.thp_morecore && __hugetlb_opts.heapbase) {
+		DEBUG("Heapbase specified with THP for morecore, ignoring heapbase\n");
+		__hugetlb_opts.heapbase = NULL;
+	}
+
 	env = getenv("HUGETLB_FORCE_ELFMAP");
 	if (env && (strcasecmp(env, "yes") == 0))
 		__hugetlb_opts.force_elfmap = 1;
@@ -302,12 +380,12 @@ void hugetlbfs_setup_env()
 
 	/* Determine if shmget() calls should be overridden */
 	env = getenv("HUGETLB_SHM");
-	if (env && !strcmp(env, "yes"))
+	if (env && !strcasecmp(env, "yes"))
 		__hugetlb_opts.shm_enabled = true;
 
 	/* Determine if all reservations should be avoided */
 	env = getenv("HUGETLB_NO_RESERVE");
-	if (env && !strcmp(env, "yes"))
+	if (env && !strcasecmp(env, "yes"))
 		__hugetlb_opts.no_reserve = true;
 }
 
@@ -471,7 +549,8 @@ void probe_default_hpage_size(void)
 			 * is purely informational in nature.
 			 */
 			char msg[] = "No mount point found for default huge " \
-				"page size. Using first available mount point.";
+				"page size. Using first available mount "
+				"point.\n";
 			if (default_overrided)
 				WARNING("%s", msg);
 			else
@@ -544,7 +623,9 @@ static void find_mounts(void)
 	char path[PATH_MAX+1];
 	char line[LINE_MAXLEN + 1];
 	char *eol;
-	int bytes, err, dummy;
+	char *match;
+	char *end;
+	int bytes;
 	off_t offset;
 
 	fd = open("/proc/mounts", O_RDONLY);
@@ -573,17 +654,22 @@ static void find_mounts(void)
 		offset = bytes - (eol + 1 - line);
 		lseek(fd, -offset, SEEK_CUR);
 
-		/*
-		 * Match only hugetlbfs filesystems.
-		 * Subtle: sscanf returns the number of input items matched
-		 * and assigned.  To force sscanf to match the literal
-		 * "hugetlbfs" string we include a 'dummy' input item
-		 * following that string.
-		 */
-		err = sscanf(line, "%*s %" stringify(PATH_MAX) "s hugetlbfs "
-			"%*s %d", path, &dummy);
-		if ((err == 2) && (hugetlbfs_test_path(path) == 1))
-			add_hugetlbfs_mount(path, 0);
+		/* Match only hugetlbfs filesystems. */
+		match = strstr(line, " hugetlbfs ");
+		if (match) {
+			match = strchr(line, '/');
+			if (!match)
+				continue;
+			end = strchr(match, ' ');
+			if (!end)
+				continue;
+
+			strncpy(path, match, end - match);
+			path[end - match] = '\0';
+			if ((hugetlbfs_test_path(path) == 1) &&
+			    !(access(path, R_OK | W_OK | X_OK)))
+				add_hugetlbfs_mount(path, 0);
+		}
 	}
 	close(fd);
 }
