@@ -377,6 +377,11 @@ def elflink_test(cmd, **env):
     do_test(cmd, **env)
     # Test we don't blow up if not linked for hugepage
     do_test(cmd, LD_PRELOAD="libhugetlbfs.so", **env)
+
+    # Only run custom ldscript tests when -l option is set
+    if not custom_ldscripts:
+        return
+
     do_elflink_test("xB." + cmd, **env)
     do_elflink_test("xBDT." + cmd, **env)
     # Test we don't blow up if HUGETLB_MINIMAL_COPY is diabled
@@ -522,15 +527,18 @@ def functional_tests():
     do_test("ptrace-write-hugepage")
     do_test("icache-hygiene")
     do_test("slbpacaflush")
-    do_test("straddle_4GB", bits=(64,))
-    do_test("huge_at_4GB_normal_below", bits=(64,))
-    do_test("huge_below_4GB_normal_above", bits=(64,))
+    do_test("straddle_4GB_static", bits=(64,))
+    do_test("huge_at_4GB_normal_below_static", bits=(64,))
+    do_test("huge_below_4GB_normal_above_static", bits=(64,))
     do_test("map_high_truncate_2")
     do_test("misaligned_offset")
     do_test("truncate_above_4GB")
     do_test("brk_near_huge")
     do_test("task-size-overrun")
     do_test_with_rlimit(resource.RLIMIT_STACK, -1, "stack_grow_into_huge")
+    do_test("corrupt-by-cow-opt")
+    do_test("noresv-preserve-resv-page")
+    do_test("noresv-regarded-as-resv")
 
     if dangerous == 1:
         do_test("readahead_reserve")
@@ -554,6 +562,10 @@ def functional_tests():
     do_test("direct")
     do_test("malloc")
     do_test("malloc", LD_PRELOAD="libhugetlbfs.so", HUGETLB_MORECORE="yes")
+    do_test("malloc", LD_PRELOAD="libhugetlbfs.so", HUGETLB_MORECORE="yes",
+            HUGETLB_RESTRICT_EXE="unknown:none")
+    do_test("malloc", LD_PRELOAD="libhugetlbfs.so", HUGETLB_MORECORE="yes",
+            HUGETLB_RESTRICT_EXE="unknown:malloc")
     do_test("malloc_manysmall")
     do_test("malloc_manysmall", LD_PRELOAD="libhugetlbfs.so",
             HUGETLB_MORECORE="yes")
@@ -563,7 +575,7 @@ def functional_tests():
     do_test("heapshrink", LD_PRELOAD="libhugetlbfs.so libheapshrink.so",
             HUGETLB_MORECORE="yes")
     do_test("heapshrink", LD_PRELOAD="libheapshrink.so", HUGETLB_MORECORE="yes",
-            HUGETLB_MORECORE_SHRINK="es")
+            HUGETLB_MORECORE_SHRINK="yes")
     do_test("heapshrink", LD_PRELOAD="libhugetlbfs.so libheapshrink.so",
             HUGETLB_MORECORE="yes", HUGETLB_MORECORE_SHRINK="yes")
     do_test("heap-overflow", HUGETLB_VERBOSE="1", HUGETLB_MORECORE="yes")
@@ -574,9 +586,11 @@ def functional_tests():
     elflink_test("linkhuge_nofd", HUGETLB_VERBOSE="0")
     elflink_test("linkhuge")
 
-    # Original elflink sharing tests
-    elfshare_test("linkshare")
-    elflink_and_share_test("linkhuge")
+    # Only run custom ldscript tests when -l option is set
+    if custom_ldscripts:
+        # Original elflink sharing tests
+        elfshare_test("linkshare")
+        elflink_and_share_test("linkhuge")
 
     # elflink_rw tests
     elflink_rw_test("linkhuge_rw")
@@ -597,7 +611,11 @@ def functional_tests():
 
     # Test overriding of shmget()
     do_shm_test("shmoverride_linked")
+    do_shm_test("shmoverride_linked", HUGETLB_SHM="yes")
+    do_shm_test("shmoverride_linked_static")
+    do_shm_test("shmoverride_linked_static", HUGETLB_SHM="yes")
     do_shm_test("shmoverride_unlinked", LD_PRELOAD="libhugetlbfs.so")
+    do_shm_test("shmoverride_unlinked", LD_PRELOAD="libhugetlbfs.so", HUGETLB_SHM="yes")
 
     # Test hugetlbfs filesystem quota accounting
     do_test("quota.sh")
@@ -637,15 +655,17 @@ def stress_tests():
 
 def main():
     global wordsizes, pagesizes, dangerous, paranoid_pool_check, system_default_hpage_size
+    global custom_ldscripts
     testsets = set()
     env_override = {"QUIET_TEST": "1", "HUGETLBFS_MOUNTS": "",
                     "HUGETLB_ELFMAP": None, "HUGETLB_MORECORE": None}
     env_defaults = {"HUGETLB_VERBOSE": "0"}
     dangerous = 0
     paranoid_pool_check = False
+    custom_ldscripts = False
 
     try:
-        opts, args = getopt.getopt(sys.argv[1:], "vVfdt:b:p:c")
+        opts, args = getopt.getopt(sys.argv[1:], "vVfdt:b:p:c:l")
     except getopt.GetoptError, err:
         print str(err)
         sys.exit(1)
@@ -665,6 +685,8 @@ def main():
            for p in arg.split(): pagesizes.add(int(p))
        elif opt == '-c':
            paranoid_pool_check = True
+       elif opt == '-l':
+           custom_ldscripts = True
        else:
            assert False, "unhandled option"
     if len(testsets) == 0: testsets = set(["func", "stress"])
